package bou.amine.apps.readerforselfossv2.android

import android.content.Context
import android.os.Environment.DIRECTORY_PICTURES
import android.os.Environment.getExternalStoragePublicDirectory
import android.util.Log
import androidx.annotation.ArrayRes
import androidx.test.espresso.Espresso
import androidx.test.espresso.Espresso.onData
import androidx.test.espresso.Espresso.onView
import androidx.test.espresso.action.ViewActions.click
import androidx.test.espresso.action.ViewActions.replaceText
import androidx.test.espresso.action.ViewActions.typeTextIntoFocusedView
import androidx.test.espresso.assertion.ViewAssertions.doesNotExist
import androidx.test.espresso.assertion.ViewAssertions.matches
import androidx.test.espresso.base.DefaultFailureHandler
import androidx.test.espresso.matcher.ViewMatchers.isChecked
import androidx.test.espresso.matcher.ViewMatchers.isDisplayed
import androidx.test.espresso.matcher.ViewMatchers.isNotChecked
import androidx.test.espresso.matcher.ViewMatchers.withId
import androidx.test.espresso.matcher.ViewMatchers.withText
import androidx.test.platform.app.InstrumentationRegistry.getInstrumentation
import androidx.test.runner.screenshot.BasicScreenCaptureProcessor
import androidx.test.runner.screenshot.Screenshot
import androidx.test.uiautomator.UiDevice
import androidx.test.uiautomator.UiSelector
import org.hamcrest.CoreMatchers.allOf
import org.hamcrest.Matchers.hasToString
import org.junit.BeforeClass
import org.junit.rules.TestWatcher
import org.junit.runner.Description
import java.io.File
import java.io.IOException
import java.util.Locale

val defaultUrl = (System.getenv("SELFOSS_URL") ?: "").ifEmpty { "http://10.0.2.2:8888" }

fun performLogin(someUrl: String? = null) {
    Log.i("AUTOMATION", "The url used will be ${if (!someUrl.isNullOrEmpty()) someUrl else defaultUrl}")
    onView(withId(R.id.urlView)).perform(click()).perform(
        typeTextIntoFocusedView(
            if (!someUrl.isNullOrEmpty()) someUrl else defaultUrl,
        ),
    )
    onView(withId(R.id.signInButton)).perform(click())
}

fun changeAndCancelSetting(
    oldValue: String,
    newValue: String,
    openSettingItem: () -> Unit,
) {
    openSettingItem()
    onView(
        withId(android.R.id.edit),
    ).perform(replaceText(newValue))
    onView(
        withId(android.R.id.button2),
    ).perform(click())
    openSettingItem()
    onView(
        withId(android.R.id.edit),
    ).check(matches(withText(oldValue)))
    onView(
        withText(newValue),
    ).check(doesNotExist())
    onView(
        withId(android.R.id.button2),
    ).perform(click())
}

fun changeAndSaveSetting(
    oldValue: String,
    newValue: String,
    openSettingItem: () -> Unit,
) {
    openSettingItem()
    onView(
        withId(android.R.id.edit),
    ).perform(replaceText(newValue))
    onView(
        withId(android.R.id.button1),
    ).perform(click())
    openSettingItem()
    onView(
        withId(android.R.id.edit),
    ).check(matches(withText(newValue)))
    if (oldValue.isNotEmpty()) {
        onView(
            withText(oldValue),
        ).check(doesNotExist())
    }
    onView(
        withId(android.R.id.button2),
    ).perform(click())
}

fun testPreferencesFromArray(
    context: Context,
    @ArrayRes arrayRes: Int,
    openSettingItem: () -> Unit,
) {
    openSettingItem()
    context.resources.getStringArray(arrayRes).forEach { res ->
        onView(withText(res)).check(matches(allOf(isDisplayed(), isNotChecked())))
        onView(withText(res)).perform(click())
        onView(withText(res)).check(doesNotExist())
        openSettingItem()
        onView(withText(res)).check(matches(allOf(isDisplayed(), isChecked())))
    }
}

fun goToSources() {
    openMenu()
    onView(withText(R.string.menu_home_sources))
        .perform(click())
}

fun testAddSourceWithUrl(
    url: String,
    sourceName: String,
) {
    onView(withId(R.id.fab))
        .perform(click())
    onView(withId(R.id.nameInput))
        .perform(click())
        .perform(typeTextIntoFocusedView(sourceName))
    onView(withId(R.id.sourceUri))
        .perform(click())
        .perform(typeTextIntoFocusedView(url))
    onView(withId(R.id.tags))
        .perform(click())
        .perform(typeTextIntoFocusedView("tag1,tag2,tag3"))
    onView(withId(R.id.spoutsSpinner))
        .perform(click())
    onData(hasToString("RSS Feed")).perform(click())
    onView(withId(R.id.saveBtn))
        .perform(click())
    onView(withText(sourceName)).check(matches(isDisplayed()))
}

@Suppress("detekt:UtilityClassWithPublicConstructor")
open class WithANRException {
    companion object {
        // Running count of the number of Android Not Responding dialogues to prevent endless dismissal.
        private var anrCount = 0

        // `RootViewWithoutFocusException` class is private, need to match the message (instead of using type matching).
        private val rootViewWithoutFocusExceptionMsg =
            java.lang.String.format(
                Locale.ROOT,
                "Waited for the root of the view hierarchy to have " +
                    "window focus and not request layout for 10 seconds. If you specified a non " +
                    "default root matcher, it may be picking a root that never takes focus. " +
                    "Root:",
            )

        private fun handleAnrDialogue() {
            val device = UiDevice.getInstance(getInstrumentation())
            // If running the device in English Locale
            val waitButton = device.findObject(UiSelector().textContains("wait"))
            if (waitButton.exists()) waitButton.click()
        }

        @JvmStatic
        @BeforeClass
        fun setUpHandler() {
            Espresso.setFailureHandler { error, viewMatcher ->

                if (error.message!!.contains(rootViewWithoutFocusExceptionMsg) && anrCount < 3) {
                    anrCount++
                    handleAnrDialogue()
                } else { // chain all failures down to the default espresso handler
                    DefaultFailureHandler(getInstrumentation().targetContext).handle(error, viewMatcher)
                }
            }
        }
    }
}

class MyScreenCaptureProcessor(
    parentFolderPath: String,
) : BasicScreenCaptureProcessor() {
    init {
        this.mDefaultScreenshotPath =
            File(
                File(
                    getExternalStoragePublicDirectory(DIRECTORY_PICTURES),
                    "selfoss_tests",
                ).absolutePath,
                "screenshots/$parentFolderPath",
            )
    }

    override fun getFilename(prefix: String): String = prefix
}

fun takeScreenshot(
    parentFolderPath: String = "",
    screenShotName: String,
) {
    Log.d("Screenshots", "Taking screenshot of '$screenShotName'")
    val screenCapture = Screenshot.capture()
    val processors = setOf(MyScreenCaptureProcessor(parentFolderPath))
    try {
        screenCapture.apply {
            name = screenShotName
            process(processors)
        }
        Log.d("Screenshots", "Screenshot taken")
    } catch (ex: IOException) {
        Log.e("Screenshots", "Could not take the screenshot", ex)
    }
}

class ScreenshotTakingRule : TestWatcher() {
    override fun failed(
        e: Throwable?,
        description: Description,
    ) {
        val parentFolderPath = "failures/${description.className}"
        takeScreenshot(parentFolderPath = parentFolderPath, screenShotName = description.methodName)
    }
}
